@@ -8,14 +8,17 @@ cli = typer.Typer()
 
 
 @cli.command()
-def install(mode: str = typer.Argument(..., help="Mode of the kernel to install.")):
-    display_name = f"Python 3 (akernel-{mode})"
+def install(mode: str = typer.Argument("", help="Mode of the kernel to install.")):
+    name = "akernel"
+    if mode:
+        name += f"-{mode}"
+    display_name = f"Python 3 ({name})"
     write_kernelspec("akernel", mode, display_name)
 
 
 @cli.command()
 def launch(
-    mode: str = typer.Argument(..., help="Mode of the kernel to launch."),
+    mode: str = typer.Argument("", help="Mode of the kernel to launch."),
     connection_file: str = typer.Option(..., "-f", help="Path to the connection file."),
 ):
     Kernel(mode, connection_file)
